import hashlib
import inspect
import time
from typing import Any, Callable, Tuple


def getRunTime(times: int = 1, output: bool = False) -> Callable:
    """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

    装饰后的函数将返回一个元组:(原函数返回值, 运行时长)
    运行时长: 类型int, 单位ns

    :param times: 运行次数, 默认为1
    :param output: 是否输出到流, 默认为False
    :return: 装饰后的函数
    """

    def ruturnFun(fun: Callable) -> Callable:
        """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

        :param fun: 原函数
        :return: 装饰器函数
        """

        def runFun(*args, **kwargs) -> Tuple[Any, int]:
            """一个装饰器, 用来计算函数运行时长, 第一个装饰器参数是运行次数

            :return: 返回一个元组:(原函数调用结果, ns为单位的运行时长)
            """
            _time_start = time.perf_counter_ns()
            for _ in range(times):
                fun_ret = fun(*args, **kwargs)
            time_used = time.perf_counter_ns() - _time_start
            if output:
                print(time_used)
            return fun_ret, time_used

        return runFun

    return ruturnFun


reRunTimes = getRunTime


def getFunName(fun: Callable) -> Callable:
    """函数形参增加__fun_name__用于获得函数名

    :param fun: 原函数
    :return: 装饰后的函数
    """
    name = fun.__name__

    def ret_fun(*args, **kwargs) -> Any:
        """装饰函数

        :return: 函数运算结果
        """
        return fun(*args, __fun_name__=name, **kwargs)

    return ret_fun


funName = getFunName


def isChange(
    hash: str = "0", ignore_line: int = 1, show_hash: bool = False
) -> Callable:
    """一个装饰器, 用来计算函数是否被修改

    :param hash: 预先计算的hash值, 默认为"0"
    :param ignore_line: 忽略函数的前几行不进行hash计算, 默认为1
    :param show_hash: 是否输出本次计算的hash值, 默认为False
    :return: 装饰后的函数
    """

    def ruturnFun(fun: Callable) -> Callable:
        """一个装饰器, 用来计算函数是否被修改

        :param fun: 原函数
        :return: 装饰器函数
        """

        def runFun(*args, **kwargs) -> tuple[Any, bool]:
            """一个装饰器, 用来计算函数是否被修改

            :return: 返回一个元组, 第一项为原返回值, 第二项是是否被修改, 被修改为True, 反之为False
            """
            is_change: bool = (
                False
                if (
                    ret_check_code := hashlib.sha224(
                        bytes(
                            "".join(
                                inspect.getsource(fun).split("\n")[ignore_line:]
                            ).encode("utf-8")
                        )
                    ).hexdigest()
                )
                == hash
                else True
            )  # 获取函数源码 去掉第一行装饰器 再获取hash 最后检查计算出来的hash和输入的hash是否一致
            if show_hash:
                print(ret_check_code)

            fun_ret = fun(*args, **kwargs)

            return fun_ret, is_change

        return runFun

    return ruturnFun



from typing import Union
from array import ArrayType


def flatten(sequence: Union[list, tuple]) -> list:
    """将多维数据结构展平为一纬数据结构

    :param sequence: 多维数据结构
    :return: 一纬数据结构
    """
    ret = []
    for _ in sequence:
        if isinstance(_, list):
            ret.extend(flatten(_))
        else:
            ret.append(_)

    # [(flatten(_) if isinstance(_, list) else _) for _ in sequence]
    return ret


def flatten_layer(sequence: Union[list, tuple]) -> list:
    """将多维数据结构展平一层

    :param sequence: 多维数据结构
    :return: 展平了一层的一纬数据结构
    """
    ret = []
    for _ in sequence:
        ret.extend(_ if isinstance(_, list) else [_])
    return ret


def flatten_no_recursion(sequence: Union[list, tuple]) -> Union[list, tuple]:
    """将多维数据结构展平为一纬数据结构(无递归)

    :param sequence: 多维数据结构
    :return: 一纬数据结构
    """
    while True:
        if all(not isinstance(_, list) for _ in sequence):  # 全不是列表
            return sequence
        sequence = flatten_layer(sequence)


def expand_dims(
    array: Union[ArrayType, list], *dims, padding_value: Union[str, int, float] = 0
):
    """将一纬数据结构提升至多维

    :param array: 一维列表/数组
    :param dims: 各项维度上限 (顺序从最里层向外)
    :param padding_value: 填充值 默认为0
    """
    allLen: int = 1  # 总长
    for num in dims:
        allLen *= num
    if len(array) < allLen:  # 补齐长度
        array.extend([padding_value for _ in range(allLen - len(array))])

    for n in dims:
        array = [array[i : i + n] for i in range(0, len(array), n)]
    return array[0]



"""默认导入hpycore的所有属性和函数，易于使用"""
from .hpycore import (
    write,
    write_without_flush,
    flush,
    output,
    setIoInstance,
    getIoInstance,
    _message_queue,
    STRING,
    NUM,
    FLOAT,
    ON,
    OFF,
    RETURN_ONCE,
    RETURN_ITERABLE,
    NO_RETURN,
    NO_RETURN_SINGLE_FUNCTION,
)
from .hpysettings import SettingsFileObject

name = "hpyculator"
__all__ = ["hpy" + body for body in "core, decorator, settings, func".split(",")]



from random import randint

from . import hpydecorator

# 测试数据
num = randint(0, 99999)
test_data = str(num) + "test_data"


def test_hpydecorator():
    test_times: int = 0  # 初始化一个变量，用于检测结果

    @hpydecorator.getRunTime(5)
    def _fun(text: str):
        nonlocal test_times
        test_times += 1
        return text

    ret = _fun(test_data)
    assert ret[0] == test_data and isinstance(ret[1], int)

    @hpydecorator.getRunTime(
        5,
    )
    def _fun(text: str):
        nonlocal test_times
        test_times += 1
        return text

    ret = _fun(test_data)

    @hpydecorator.getFunName
    def _fun2(text: str, __fun_name__: str):
        return text, __fun_name__

    assert (test_data, "_fun2") == _fun2(test_data)

    @hpydecorator.isChange(
        hash="0",
        ignore_line=4,
        show_hash=True,
    )
    def _fun3() -> None:
        """func

        :return: None
        """
        print("hello")

    assert _fun3()[1] is True  # 代码被修改

    @hpydecorator.isChange(
        hash="5b0ad0d14da3144eb78b73c950320b145d3939d3ef60aa4d52616989",
        ignore_line=4,
        show_hash=False,
    )
    def _fun4() -> None:
        """func

        :return: None
        """
        print("hello")

    assert _fun4()[1] is False  # 代码未被修改
